@@ -48,7 +48,7 @@ public:
         return true;
     }
 
-    bool aboutShowCopyright() const {
+    bool aboutShowCopyright() const override {
         return false;
     }
 
@@ -64,6 +64,10 @@ public:
         return QStringLiteral("https://www.transip.nl/support");
     }
 
+    QString defaultClientFolder() const {
+        return QStringLiteral("STACK");
+    }
+
     QString wizardUrlHint() const override {
         return QStringLiteral("https://example.stack.storage");
     }
@@ -88,7 +92,7 @@ public:
             "United States, other countries, or both.</p>"
         )
         .arg(
-            Utility::escape(version()),
+            Utility::escape(OCC::Version::displayString()),
             Utility::escape(QStringLiteral("https://" APPLICATION_DOMAIN)),
             Utility::escape(QStringLiteral(APPLICATION_DOMAIN)),
             Utility::escape(QStringLiteral("https://www.transip.nl/knowledgebase/artikel/5558-changelog-van-de-desktopapplicatie/")),
