@@ -17,21 +17,18 @@
  */
 
 #include "filesystembase.h"
+#include "utility.h"
+#include "common/asserts.h"
 
 #include <QDateTime>
 #include <QDir>
 #include <QUrl>
 #include <QFile>
-#include <QCryptographicHash>
 #include <QCoreApplication>
 
 #include <sys/stat.h>
 #include <sys/types.h>
 
-#ifdef ZLIB_FOUND
-#include <zlib.h>
-#endif
-
 #ifdef Q_OS_WIN
 #include <windows.h>
 #include <windef.h>
@@ -40,16 +37,43 @@
 #include <io.h>
 #endif
 
+namespace {
+// Regarding
+// https://en.wikipedia.org/wiki/Comparison_of_file_systems#Limits
+// the file name limit appears to be about 255 chars
+// -10 for safety
+constexpr int fileNameMaxC = 255 - 10;
+
+constexpr auto replacementCharC = QLatin1Char('_');
+}
+
 namespace OCC {
 
-Q_LOGGING_CATEGORY(lcFileSystem, "nextcloud.sync.filesystem", QtInfoMsg)
+Q_LOGGING_CATEGORY(lcFileSystem, "sync.filesystem", QtInfoMsg)
 
 QString FileSystem::longWinPath(const QString &inpath)
 {
-#ifdef Q_OS_WIN
-    return pathtoUNC(inpath);
-#else
+#ifndef Q_OS_WIN
     return inpath;
+#else
+    Q_ASSERT(QFileInfo(inpath).isAbsolute());
+    if (inpath.isEmpty()) {
+        return inpath;
+    }
+    const QString str = QDir::toNativeSeparators(inpath);
+    const QLatin1Char sep('\\');
+
+    // we already have a unc path
+    if (str.startsWith(sep + sep)) {
+        return str;
+    }
+    // prepend \\?\ and to support long names
+
+    if (str.at(0) == sep) {
+        // should not happen as we require the path to be absolute
+        return QStringLiteral("\\\\?") + str;
+    }
+    return QStringLiteral("\\\\?\\") + str;
 #endif
 }
 
@@ -103,6 +127,7 @@ void FileSystem::setFileReadOnly(const QString &filename, bool readonly)
     if (!readonly) {
         permissions |= defaultWritePermissions;
     }
+    // Warning: This function does not manipulate ACLs, which may limit its effectiveness.
     file.setPermissions(permissions);
 }
 
@@ -137,21 +162,18 @@ bool FileSystem::rename(const QString &originFileName,
     bool success = false;
     QString error;
 #ifdef Q_OS_WIN
-    QString orig = longWinPath(originFileName);
-    QString dest = longWinPath(destinationFileName);
-
-    if (isLnkFile(originFileName) || isLnkFile(destinationFileName)) {
+    const QString orig = longWinPath(originFileName);
+    const QString dest = longWinPath(destinationFileName);
+    if (FileSystem::isFileLocked(dest, FileSystem::LockMode::Exclusive)) {
+        error = QCoreApplication::translate("FileSystem", "Can't rename %1, the file is currently in use").arg(destinationFileName);
+    } else if (FileSystem::isFileLocked(orig, FileSystem::LockMode::Exclusive)) {
+        error = QCoreApplication::translate("FileSystem", "Can't rename %1, the file is currently in use").arg(originFileName);
+    } else if (isLnkFile(originFileName) || isLnkFile(destinationFileName)) {
         success = MoveFileEx((wchar_t *)orig.utf16(),
             (wchar_t *)dest.utf16(),
             MOVEFILE_COPY_ALLOWED | MOVEFILE_WRITE_THROUGH);
         if (!success) {
-            wchar_t *string = 0;
-            FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
-                nullptr, ::GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
-                (LPWSTR)&string, 0, nullptr);
-
-            error = QString::fromWCharArray(string);
-            LocalFree((HLOCAL)string);
+            error = Utility::formatWinError(GetLastError());
         }
     } else
 #endif
@@ -178,6 +200,7 @@ bool FileSystem::uncheckedRenameReplace(const QString &originFileName,
     const QString &destinationFileName,
     QString *errorString)
 {
+    Q_ASSERT(errorString);
 #ifndef Q_OS_WIN
     bool success;
     QFile orig(originFileName);
@@ -202,26 +225,28 @@ bool FileSystem::uncheckedRenameReplace(const QString &originFileName,
 
 #else //Q_OS_WIN
     // You can not overwrite a read-only file on windows.
+
     if (!QFileInfo(destinationFileName).isWritable()) {
         setFileReadOnly(destinationFileName, false);
     }
-
-    BOOL ok;
-    QString orig = longWinPath(originFileName);
-    QString dest = longWinPath(destinationFileName);
-
-    ok = MoveFileEx((wchar_t *)orig.utf16(),
-        (wchar_t *)dest.utf16(),
-        MOVEFILE_REPLACE_EXISTING + MOVEFILE_COPY_ALLOWED + MOVEFILE_WRITE_THROUGH);
+    const QString orig = longWinPath(originFileName);
+    const QString dest = longWinPath(destinationFileName);
+    if (FileSystem::isFileLocked(dest, FileSystem::LockMode::Exclusive)) {
+        *errorString = QCoreApplication::translate("FileSystem", "Can't rename %1, the file is currently in use").arg(destinationFileName);
+        qCWarning(lcFileSystem) << "Renaming failed: " << *errorString;
+        return false;
+    }
+    if (FileSystem::isFileLocked(orig, FileSystem::LockMode::Exclusive)) {
+        *errorString = QCoreApplication::translate("FileSystem", "Can't rename %1, the file is currently in use").arg(originFileName);
+        qCWarning(lcFileSystem) << "Renaming failed: " << *errorString;
+        return false;
+    }
+    const BOOL ok = MoveFileEx(reinterpret_cast<const wchar_t *>(orig.utf16()),
+        reinterpret_cast<const wchar_t *>(dest.utf16()),
+        MOVEFILE_REPLACE_EXISTING | MOVEFILE_COPY_ALLOWED | MOVEFILE_WRITE_THROUGH);
     if (!ok) {
-        wchar_t *string = 0;
-        FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
-            nullptr, ::GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
-            (LPWSTR)&string, 0, nullptr);
-
-        *errorString = QString::fromWCharArray(string);
+        *errorString = Utility::formatWinError(GetLastError());
         qCWarning(lcFileSystem) << "Renaming temp file to final failed: " << *errorString;
-        LocalFree((HLOCAL)string);
         return false;
     }
 #endif
@@ -248,7 +273,7 @@ bool FileSystem::openAndSeekFileSharedRead(QFile *file, QString *errorOrNull, qi
     DWORD creationDisp = OPEN_EXISTING;
 
     // Create the file handle.
-    SECURITY_ATTRIBUTES securityAtts = { sizeof(SECURITY_ATTRIBUTES), nullptr, FALSE };
+    SECURITY_ATTRIBUTES securityAtts = { sizeof(SECURITY_ATTRIBUTES), NULL, FALSE };
     QString fName = longWinPath(file->fileName());
 
     HANDLE fileHandle = CreateFileW(
@@ -258,7 +283,7 @@ bool FileSystem::openAndSeekFileSharedRead(QFile *file, QString *errorOrNull, qi
         &securityAtts,
         creationDisp,
         FILE_ATTRIBUTE_NORMAL,
-        nullptr);
+        NULL);
 
     // Bail out on error.
     if (fileHandle == INVALID_HANDLE_VALUE) {
@@ -271,7 +296,7 @@ bool FileSystem::openAndSeekFileSharedRead(QFile *file, QString *errorOrNull, qi
     // the fd the handle will be closed too.
     int fd = _open_osfhandle((intptr_t)fileHandle, _O_RDONLY);
     if (fd == -1) {
-        error = "could not make fd from handle";
+        error = QStringLiteral("could not make fd from handle");
         CloseHandle(fileHandle);
         return false;
     }
@@ -332,8 +357,7 @@ bool FileSystem::fileExists(const QString &filename, const QFileInfo &fileInfo)
     // not valid. There needs to be one initialised here. Otherwise the incoming
     // fileInfo is re-used.
     if (fileInfo.filePath() != filename) {
-        QFileInfo myFI(filename);
-        re = myFI.exists();
+        re = QFileInfo::exists(filename);
     }
     return re;
 }
@@ -343,17 +367,17 @@ QString FileSystem::fileSystemForPath(const QString &path)
 {
     // See also QStorageInfo (Qt >=5.4) and GetVolumeInformationByHandleW (>= Vista)
     QString drive = path.left(2);
-    if (!drive.endsWith(":"))
+    if (!drive.endsWith(QLatin1Char(':')))
         return QString();
-    drive.append('\\');
+    drive.append(QLatin1Char('\\'));
 
     const size_t fileSystemBufferSize = 4096;
     TCHAR fileSystemBuffer[fileSystemBufferSize];
 
     if (!GetVolumeInformationW(
             reinterpret_cast<LPCWSTR>(drive.utf16()),
-            nullptr, 0,
-            nullptr, nullptr, nullptr,
+            NULL, 0,
+            NULL, NULL, NULL,
             fileSystemBuffer, fileSystemBufferSize)) {
         return QString();
     }
@@ -361,58 +385,12 @@ QString FileSystem::fileSystemForPath(const QString &path)
 }
 #endif
 
-#define BUFSIZE qint64(500 * 1024) // 500 KiB
-
-static QByteArray readToCrypto( const QString& filename, QCryptographicHash::Algorithm algo )
- {
-     QFile file(filename);
-     QByteArray arr;
-     QCryptographicHash crypto( algo );
-
-     if (file.open(QIODevice::ReadOnly)) {
-         if (crypto.addData(&file)) {
-             arr = crypto.result().toHex();
-         }
-     }
-     return arr;
- }
-
-QByteArray FileSystem::calcMd5(const QString &filename)
-{
-    return readToCrypto(filename, QCryptographicHash::Md5);
-}
-
-QByteArray FileSystem::calcSha1(const QString &filename)
-{
-    return readToCrypto(filename, QCryptographicHash::Sha1);
-}
-
-#ifdef ZLIB_FOUND
-QByteArray FileSystem::calcAdler32(const QString &filename)
-{
-    QFile file(filename);
-    const qint64 bufSize = qMin(BUFSIZE, file.size() + 1);
-    QByteArray buf(bufSize, Qt::Uninitialized);
-
-    unsigned int adler = adler32(0L, Z_NULL, 0);
-    if (file.open(QIODevice::ReadOnly)) {
-        qint64 size;
-        while (!file.atEnd()) {
-            size = file.read(buf.data(), bufSize);
-            if (size > 0)
-                adler = adler32(adler, (const Bytef *)buf.data(), size);
-        }
-    }
-
-    return QByteArray::number(adler, 16);
-}
-#endif
-
 bool FileSystem::remove(const QString &fileName, QString *errorString)
 {
 #ifdef Q_OS_WIN
     // You cannot delete a read-only file on windows, but we want to
     // allow that.
+
     if (!QFileInfo(fileName).isWritable()) {
         setFileReadOnly(fileName, false);
     }
@@ -422,6 +400,7 @@ bool FileSystem::remove(const QString &fileName, QString *errorString)
         if (errorString) {
             *errorString = f.errorString();
         }
+        qCWarning(lcFileSystem) << "Failed to delete:" << fileName << "Error:" << f.errorString();
         return false;
     }
     return true;
@@ -429,17 +408,18 @@ bool FileSystem::remove(const QString &fileName, QString *errorString)
 
 bool FileSystem::moveToTrash(const QString &fileName, QString *errorString)
 {
+    // TODO: Qt 5.15 bool QFile::moveToTrash()
 #if defined Q_OS_UNIX && !defined Q_OS_MAC
     QString trashPath, trashFilePath, trashInfoPath;
     QString xdgDataHome = QFile::decodeName(qgetenv("XDG_DATA_HOME"));
     if (xdgDataHome.isEmpty()) {
-        trashPath = QDir::homePath() + "/.local/share/Trash/"; // trash path that should exist
+        trashPath = QDir::homePath() + QStringLiteral("/.local/share/Trash/"); // trash path that should exist
     } else {
-        trashPath = xdgDataHome + "/Trash/";
+        trashPath = xdgDataHome + QStringLiteral("/Trash/");
     }
 
-    trashFilePath = trashPath + "files/"; // trash file path contain delete files
-    trashInfoPath = trashPath + "info/"; // trash info path contain delete files information
+    trashFilePath = trashPath + QStringLiteral("files/"); // trash file path contain delete files
+    trashInfoPath = trashPath + QStringLiteral("info/"); // trash info path contain delete files information
 
     if (!(QDir().mkpath(trashFilePath) && QDir().mkpath(trashInfoPath))) {
         *errorString = QCoreApplication::translate("FileSystem", "Could not make directories in trash");
@@ -451,7 +431,7 @@ bool FileSystem::moveToTrash(const QString &fileName, QString *errorString)
     QDir file;
     int suffix_number = 1;
     if (file.exists(trashFilePath + f.fileName())) { //file in trash already exists, move to "filename.1"
-        QString path = trashFilePath + f.fileName() + ".";
+        QString path = trashFilePath + f.fileName() + QLatin1Char('.');
         while (file.exists(path + QString::number(suffix_number))) { //or to "filename.2" if "filename.1" exists, etc
             suffix_number++;
         }
@@ -470,11 +450,11 @@ bool FileSystem::moveToTrash(const QString &fileName, QString *errorString)
 
     // create file format for trash info file----- START
     QFile infoFile;
-    if (file.exists(trashInfoPath + f.fileName() + ".trashinfo")) { //TrashInfo file already exists, create "filename.1.trashinfo"
-        QString filename = trashInfoPath + f.fileName() + "." + QString::number(suffix_number) + ".trashinfo";
+    if (file.exists(trashInfoPath + f.fileName() + QStringLiteral(".trashinfo"))) { //TrashInfo file already exists, create "filename.1.trashinfo"
+        QString filename = trashInfoPath + f.fileName() + QLatin1Char('.') + QString::number(suffix_number) + QStringLiteral(".trashinfo");
         infoFile.setFileName(filename); //filename+.trashinfo //  create file information file in /.local/share/Trash/info/ folder
     } else {
-        QString filename = trashInfoPath + f.fileName() + ".trashinfo";
+        QString filename = trashInfoPath + f.fileName() + QStringLiteral(".trashinfo");
         infoFile.setFileName(filename); //filename+.trashinfo //  create file information file in /.local/share/Trash/info/ folder
     }
 
@@ -482,16 +462,13 @@ bool FileSystem::moveToTrash(const QString &fileName, QString *errorString)
 
     QTextStream stream(&infoFile); // for write data on open file
 
-    QByteArray info = "[Trash Info]\n";
-    info += "Path=";
-    info += QUrl::toPercentEncoding(f.absoluteFilePath(), "~_-./");
-    info += '\n';
-    info += "DeletionDate=";
-    info += QDateTime::currentDateTime().toString(Qt::ISODate).toLatin1();
-    info += '\n';
-
-    stream << info;
-
+    stream << "[Trash Info]\n"
+           << "Path="
+           << QUrl::toPercentEncoding(f.absoluteFilePath(), "~_-./")
+           << "\n"
+           << "DeletionDate="
+           << QDateTime::currentDateTime().toString(Qt::ISODate)
+           << '\n';
     infoFile.close();
 
     // create info file format of trash file----- END
@@ -504,46 +481,57 @@ bool FileSystem::moveToTrash(const QString &fileName, QString *errorString)
 #endif
 }
 
-bool FileSystem::isFileLocked(const QString &fileName)
-{
 #ifdef Q_OS_WIN
-    const wchar_t *wuri = reinterpret_cast<const wchar_t *>(fileName.utf16());
+Utility::Handle FileSystem::lockFile(const QString &fileName, LockMode mode)
+{
+    const QString fName = longWinPath(fileName);
+    auto accessMode = mode == LockMode::Exclusive ? 0 : FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE;
     // Check if file exists
-    DWORD attr = GetFileAttributesW(wuri);
+    DWORD attr = GetFileAttributesW(reinterpret_cast<const wchar_t *>(fName.utf16()));
     if (attr != INVALID_FILE_ATTRIBUTES) {
         // Try to open the file with as much access as possible..
-        HANDLE win_h = CreateFileW(
-            wuri,
+        return Utility::Handle { CreateFileW(
+            reinterpret_cast<const wchar_t *>(fName.utf16()),
             GENERIC_READ | GENERIC_WRITE,
-            FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
+            accessMode,
             nullptr, OPEN_EXISTING,
             FILE_ATTRIBUTE_NORMAL | FILE_FLAG_BACKUP_SEMANTICS,
-            nullptr);
+            nullptr) };
+    }
+    return {};
+}
+#endif
 
-        if (win_h == INVALID_HANDLE_VALUE) {
-            /* could not be opened, so locked? */
-            /* 32 == ERROR_SHARING_VIOLATION */
+
+bool FileSystem::isFileLocked(const QString &fileName, LockMode mode)
+{
+#ifdef Q_OS_WIN
+    const auto handle = lockFile(fileName, mode);
+    if (!handle) {
+        const auto error = GetLastError();
+        if (error == ERROR_SHARING_VIOLATION) {
             return true;
-        } else {
-            CloseHandle(win_h);
+        } else if (error != ERROR_FILE_NOT_FOUND && error != ERROR_PATH_NOT_FOUND) {
+            qCWarning(lcFileSystem()) << Q_FUNC_INFO << Utility::formatWinError(error);
         }
     }
 #else
     Q_UNUSED(fileName);
+    Q_UNUSED(mode);
 #endif
     return false;
 }
 
 bool FileSystem::isLnkFile(const QString &filename)
 {
-    return filename.endsWith(".lnk");
+    return filename.endsWith(QLatin1String(".lnk"));
 }
 
 bool FileSystem::isJunction(const QString &filename)
 {
 #ifdef Q_OS_WIN
     WIN32_FIND_DATA findData;
-    HANDLE hFind = FindFirstFileEx((const wchar_t *)filename.utf16(), FindExInfoBasic, &findData, FindExSearchNameMatch, nullptr, 0);
+    HANDLE hFind = FindFirstFileEx(reinterpret_cast<const wchar_t *>(longWinPath(filename).utf16()), FindExInfoBasic, &findData, FindExSearchNameMatch, NULL, 0);
     if (hFind != INVALID_HANDLE_VALUE) {
         FindClose(hFind);
         return false;
@@ -557,4 +545,29 @@ bool FileSystem::isJunction(const QString &filename)
 #endif
 }
 
+bool FileSystem::isChildPathOf(const QString &child, const QString &parent)
+{
+    // if it is a relative path assume a local file, resolve it based on root
+    const auto sensitivity = Utility::fsCaseSensitivity();
+    return (child.startsWith(parent.endsWith(QLatin1Char('/')) ? parent : parent + QLatin1Char('/'), sensitivity)
+        // clear trailing slashes etc
+        || QString::compare(QDir::cleanPath(parent), QDir::cleanPath(child), sensitivity) == 0);
+}
+
+QString OCSYNC_EXPORT FileSystem::createPortableFileName(const QFileInfo &path, int reservedSize)
+{
+    // remove leading and trailing whitespace
+    QString tmp = path.fileName().trimmed();
+    // limit size to fileNameMaxC
+    tmp.resize(std::min(tmp.size(), fileNameMaxC - reservedSize));
+    // remove eventual trailing whitespace after the resize
+    tmp = tmp.trimmed();
+    // replace non portable characters
+    for (auto c : IllegalFilenameCharsWindows) {
+        tmp.replace(c, replacementCharC);
+    }
+    return path.dir().filePath(tmp);
+}
 } // namespace OCC
+
+#include "moc_filesystembase.cpp"
