@@ -12,25 +12,28 @@
  * for more details.
  */
 
-#include "application.h"
 #include "owncloudgui.h"
-#include "theme.h"
-#include "folderman.h"
-#include "configfile.h"
-#include "progressdispatcher.h"
-#include "owncloudsetupwizard.h"
-#include "sharedialog.h"
-#include "settingsdialog.h"
-#include "logger.h"
-#include "logbrowser.h"
-#include "account.h"
-#include "accountstate.h"
-#include "openfilemanager.h"
-#include "accountmanager.h"
 #include "aboutdialog.h"
+#include "account.h"
+#include "accountmanager.h"
+#include "accountstate.h"
+#include "application.h"
 #include "common/syncjournalfilerecord.h"
+#include "configfile.h"
 #include "creds/abstractcredentials.h"
+#include "folderman.h"
+#include "folderwizard/folderwizard.h"
+#include "graphapi/drives.h"
+#include "gui/accountsettings.h"
 #include "guiutility.h"
+#include "logbrowser.h"
+#include "logger.h"
+#include "openfilemanager.h"
+#include "progressdispatcher.h"
+#include "settingsdialog.h"
+#include "setupwizardcontroller.h"
+#include "sharedialog.h"
+#include "theme.h"
 
 #include <QDesktopServices>
 #include <QDir>
@@ -41,8 +44,68 @@
 
 #ifdef WITH_LIBCLOUDPROVIDERS
 #include "libcloudproviders/libcloudproviders.h"
+#include "selectivesyncdialog.h"
 #endif
 
+using namespace std::chrono_literals;
+
+namespace {
+
+using namespace OCC;
+
+void setUpInitialSyncFolder(AccountStatePtr accountStatePtr, bool useVfs)
+{
+    auto folderMan = FolderMan::instance();
+
+    // saves a bit of duplicate code
+    auto addFolder = [folderMan, accountStatePtr, useVfs](const QString &localFolder, const QString &remotePath, const QUrl &webDavUrl, const QString &displayName = {}) {
+        return folderMan->addFolderFromWizard(accountStatePtr, localFolder, remotePath, webDavUrl, displayName, useVfs);
+    };
+
+    auto finalize = [accountStatePtr] {
+        accountStatePtr->checkConnectivity();
+        FolderMan::instance()->setSyncEnabled(true);
+        FolderMan::instance()->scheduleAllFolders();
+    };
+
+    if (accountStatePtr->supportsSpaces()) {
+        auto *drive = new GraphApi::Drives(accountStatePtr->account());
+
+        QObject::connect(drive, &GraphApi::Drives::finishedSignal, [accountStatePtr, drive, addFolder, finalize] {
+            if (drive->parseError().error == QJsonParseError::NoError) {
+                auto drives = drive->drives();
+
+                // we do not want to set up folder sync connections for disabled spaces (#10173)
+                drives.erase(std::remove_if(drives.begin(), drives.end(), &GraphApi::isDriveDisabled), drives.end());
+
+                if (!drives.isEmpty()) {
+                    const QDir localDir(accountStatePtr->account()->defaultSyncRoot());
+                    FileSystem::setFolderMinimumPermissions(localDir.path());
+                    Folder::prepareFolder(localDir.path());
+                    Utility::setupFavLink(localDir.path());
+                    for (const auto &d : drives) {
+                        const QString name = GraphApi::Drives::getDriveDisplayName(d);
+                        const QString folderName = FolderMan::instance()->findGoodPathForNewSyncFolder(localDir.filePath(name));
+                        auto folder = addFolder(folderName, {}, QUrl::fromEncoded(d.getRoot().getWebDavUrl().toUtf8()), name);
+                        folder->setPriority(GraphApi::Drives::getDrivePriority(d));
+                        // save the new priority
+                        folder->saveToSettings();
+                    }
+                    finalize();
+                }
+            }
+        });
+
+        drive->start();
+
+        return;
+    } else {
+        addFolder(accountStatePtr->account()->defaultSyncRoot(), Theme::instance()->defaultServerFolder(), accountStatePtr->account()->davUrl());
+        finalize();
+    }
+}
+}
+
 namespace OCC {
 
 const char propertyAccountC[] = "oc_account";
@@ -146,13 +209,6 @@ void ownCloudGui::slotSyncStateChange(Folder *folder)
 
     qCInfo(lcApplication) << "Sync state changed for folder " << folder->remoteUrl().toString() << ": " << result.statusString();
 
-    if (result.status() == SyncResult::Success
-        || result.status() == SyncResult::Problem
-        || result.status() == SyncResult::SyncAbortRequested
-        || result.status() == SyncResult::Error) {
-        Logger::instance()->enterNextLogFile();
-    }
-
     if (result.status() == SyncResult::NotYetStarted) {
         _settingsDialog->slotRefreshActivity(folder->accountState());
     }
@@ -177,13 +233,13 @@ void ownCloudGui::slotAccountStateChanged()
 
 void ownCloudGui::slotTrayMessageIfServerUnsupported(Account *account)
 {
-    if (account->serverVersionUnsupported()) {
+    if (account->serverSupportLevel() != Account::ServerSupportLevel::Supported) {
         slotShowTrayMessage(
             tr("Unsupported Server Version"),
             tr("The server on account %1 runs an unsupported version %2. "
                "Using this client with unsupported server versions is untested and "
                "potentially dangerous. Proceed at your own risk.")
-                .arg(account->displayName(), account->serverVersion()));
+                .arg(account->displayName(), account->capabilities().status().versionString()));
     }
 }
 
@@ -211,7 +267,7 @@ void ownCloudGui::slotComputeOverallSyncStatus()
         }
     }
 
-    const auto &map = FolderMan::instance()->map();
+    const auto &map = FolderMan::instance()->folders();
     for (auto *f : map) {
         if (!f->syncPaused()) {
             allPaused = false;
@@ -263,7 +319,7 @@ void ownCloudGui::slotComputeOverallSyncStatus()
     // display the info of the least successful sync (eg. do not just display the result of the latest sync)
     QString trayMessage;
 
-    auto trayOverallStatusResult = FolderMan::trayOverallStatus(map.values());
+    auto trayOverallStatusResult = FolderMan::trayOverallStatus(map);
 
     const QIcon statusIcon = Theme::instance()->syncStateIcon(trayOverallStatusResult.overallStatus(), true, contextMenuVisible());
     _tray->setIcon(statusIcon);
@@ -324,7 +380,7 @@ void ownCloudGui::addAccountContextMenu(AccountStatePtr accountState, QMenu *men
 
     FolderMan *folderMan = FolderMan::instance();
     bool firstFolder = true;
-    const auto &map = folderMan->map();
+    const auto &map = folderMan->folders();
     bool singleSyncFolder = map.size() == 1 && Theme::instance()->singleSyncFolder();
     bool onePaused = false;
     bool allPaused = true;
@@ -346,8 +402,7 @@ void ownCloudGui::addAccountContextMenu(AccountStatePtr accountState, QMenu *men
         }
 
         QAction *action = menu->addAction(tr("Open folder '%1'").arg(folder->shortGuiLocalPath()));
-        auto alias = folder->alias();
-        connect(action, &QAction::triggered, this, [this, alias] { this->slotFolderOpenAction(alias); });
+        connect(action, &QAction::triggered, this, [this, folder] { this->slotFolderOpenAction(folder); });
     }
 
     menu->addSeparator();
@@ -525,7 +580,7 @@ void ownCloudGui::setupContextMenu()
         || desktopSession.contains("plasma")
         || desktopSession.contains("kde");
     QObject *platformMenu = reinterpret_cast<QObject *>(_tray->contextMenu()->platformMenu());
-    if (isKde && platformMenu && platformMenu->metaObject()->className() == QLatin1String("QDBusPlatformMenu")) {
+    if (isKde && platformMenu && platformMenu->metaObject()->className() == QByteArrayLiteral("QDBusPlatformMenu")) {
         _workaroundManualVisibility = true;
         _workaroundNoAboutToShowUpdate = true;
     }
@@ -544,7 +599,7 @@ void ownCloudGui::setupContextMenu()
 
 
     connect(&_delayedTrayUpdateTimer, &QTimer::timeout, this, &ownCloudGui::updateContextMenu);
-    _delayedTrayUpdateTimer.setInterval(2 * 1000);
+    _delayedTrayUpdateTimer.setInterval(2s);
     _delayedTrayUpdateTimer.setSingleShot(true);
 
     connect(_contextMenu.data(), &QMenu::aboutToShow, this, &ownCloudGui::slotContextMenuAboutToShow);
@@ -611,7 +666,7 @@ void ownCloudGui::updateContextMenu()
         }
     }
 
-    for (auto *f : FolderMan::instance()->map()) {
+    for (auto *f : FolderMan::instance()->folders()) {
         if (f->syncPaused()) {
             atLeastOnePaused = true;
         } else {
@@ -720,9 +775,8 @@ void ownCloudGui::slotShowOptionalTrayMessage(const QString &title, const QStrin
 /*
  * open the folder with the given Alias
  */
-void ownCloudGui::slotFolderOpenAction(const QString &alias)
+void ownCloudGui::slotFolderOpenAction(Folder *f)
 {
-    Folder *f = FolderMan::instance()->folder(alias);
     if (f) {
         qCInfo(lcApplication) << "opening local url " << f->path();
         QUrl url = QUrl::fromLocalFile(f->path());
@@ -744,7 +798,7 @@ void ownCloudGui::setupActions()
 {
     _actionStatus = new QAction(tr("Unknown status"), this);
     _actionStatus->setEnabled(false);
-    _actionSettings = new QAction(tr("Settings..."), this);
+    _actionSettings = new QAction(tr("Show %1").arg(Theme::instance()->appNameGUI()), this);
     _actionNewAccountWizard = new QAction(tr("New account..."), this);
     _actionRecent = new QAction(tr("Details..."), this);
     _actionRecent->setEnabled(true);
@@ -765,11 +819,11 @@ void ownCloudGui::setupActions()
     connect(_actionLogout, &QAction::triggered, this, &ownCloudGui::slotLogout);
 
     if (_app->debugMode()) {
-        _actionCrash = new QAction("Crash now - Div by zero", this);
+        _actionCrash = new QAction(QStringLiteral("Crash now - Div by zero"), this);
         connect(_actionCrash, &QAction::triggered, _app, &Application::slotCrash);
-        _actionCrashEnforce = new QAction("Crash now - ENFORCE()", this);
+        _actionCrashEnforce = new QAction(QStringLiteral("Crash now - ENFORCE()"), this);
         connect(_actionCrashEnforce, &QAction::triggered, _app, &Application::slotCrashEnforce);
-        _actionCrashFatal = new QAction("Crash now - qFatal", this);
+        _actionCrashFatal = new QAction(QStringLiteral("Crash now - qFatal"), this);
         connect(_actionCrashFatal, &QAction::triggered, _app, &Application::slotCrashFatal);
     } else {
         _actionCrash = nullptr;
@@ -802,11 +856,8 @@ static bool shouldShowInRecentsMenu(const SyncFileItem &item)
         && item._instruction != CSYNC_INSTRUCTION_NONE;
 }
 
-
-void ownCloudGui::slotUpdateProgress(const QString &folder, const ProgressInfo &progress)
+void ownCloudGui::slotUpdateProgress(Folder *folder, const ProgressInfo &progress)
 {
-    Q_UNUSED(folder);
-
     if (progress.status() == ProgressInfo::Discovery) {
         if (!progress._currentDiscoveredRemoteFolder.isEmpty()) {
             _actionStatus->setText(tr("Checking for changes in remote '%1'")
@@ -816,7 +867,7 @@ void ownCloudGui::slotUpdateProgress(const QString &folder, const ProgressInfo &
                                        .arg(progress._currentDiscoveredLocalFolder));
         }
     } else if (progress.status() == ProgressInfo::Done) {
-        QTimer::singleShot(2000, this, &ownCloudGui::slotComputeOverallSyncStatus);
+        QTimer::singleShot(2s, this, &ownCloudGui::slotComputeOverallSyncStatus);
     }
     if (progress.status() != ProgressInfo::Propagation) {
         return;
@@ -860,17 +911,14 @@ void ownCloudGui::slotUpdateProgress(const QString &folder, const ProgressInfo &
         }
 
         QString kindStr = Progress::asResultString(progress._lastCompletedItem);
-        QString timeStr = QTime::currentTime().toString("hh:mm");
+        QString timeStr = QTime::currentTime().toString(QStringLiteral("hh:mm"));
         QString actionText = tr("%1 (%2, %3)").arg(progress._lastCompletedItem._file, kindStr, timeStr);
         QAction *action = new QAction(actionText, this);
-        Folder *f = FolderMan::instance()->folder(folder);
-        if (f) {
-            QString fullPath = f->path() + '/' + progress._lastCompletedItem._file;
-            if (QFile(fullPath).exists()) {
-                connect(action, &QAction::triggered, this, [this, fullPath] { this->slotOpenPath(fullPath); });
-            } else {
-                action->setEnabled(false);
-            }
+        QString fullPath = folder->path() + QLatin1Char('/') + progress._lastCompletedItem._file;
+        if (QFile(fullPath).exists()) {
+            connect(action, &QAction::triggered, this, [this, fullPath] { this->slotOpenPath(fullPath); });
+        } else {
+            action->setEnabled(false);
         }
         if (_recentItemsActions.length() > 5) {
             _recentItemsActions.takeFirst()->deleteLater();
@@ -888,7 +936,6 @@ void ownCloudGui::slotUpdateProgress(const QString &folder, const ProgressInfo &
 void ownCloudGui::slotLogin()
 {
     if (auto account = qvariant_cast<AccountStatePtr>(sender()->property(propertyAccountC))) {
-        account->account()->resetRejectedCertificates();
         account->signIn();
     } else {
         for (const auto &a : AccountManager::instance()->accounts()) {
@@ -922,27 +969,122 @@ void ownCloudGui::slotPauseAllFolders()
 
 void ownCloudGui::runNewAccountWizard()
 {
-    if (!_wizard) {
-        _wizard = new OwncloudSetupWizard(settingsDialog());
-        connect(_wizard, &OwncloudSetupWizard::ownCloudWizardDone, _wizard, &OwncloudSetupWizard::deleteLater);
-        connect(_wizard, &OwncloudSetupWizard::ownCloudWizardDone, ocApp(), &Application::slotownCloudWizardDone);
+    if (_wizardController.isNull()) {
+        // passing the settings dialog as parent makes sure the wizard will be shown above it
+        // as the settingsDialog's lifetime spans across the entire application but the dialog will live much shorter,
+        // we have to clean it up manually when finished() is emitted
+        _wizardController = new Wizard::SetupWizardController(settingsDialog());
+
+        // while the wizard is shown, new syncs are disabled
         FolderMan::instance()->setSyncEnabled(false);
-        _wizard->startWizard();
+
+        connect(_wizardController, &Wizard::SetupWizardController::finished, ocApp(),
+            [this](AccountPtr newAccount, Wizard::SyncMode syncMode, const QVariantMap &dynamicRegistrationData) {
+                // note: while the wizard is shown, we disable the folder synchronization
+                // previously we could perform this just here, but now we have to postpone this depending on whether selective sync was chosen
+                // see also #9497
+
+                // when the dialog is closed before it has finished, there won't be a new account to set up
+                // the wizard controller signalizes this by passing a null pointer
+                if (!newAccount.isNull()) {
+                    // finally, call the slot that finalizes the setup
+                    auto accountStatePtr = ocApp()->addNewAccount(newAccount);
+
+                    // ensure we are connected and fetch the capabilities
+                    auto validator = new ConnectionValidator(accountStatePtr->account(), accountStatePtr->account().data());
+
+                    QObject::connect(validator, &ConnectionValidator::connectionResult, accountStatePtr.data(), [accountStatePtr, syncMode, dynamicRegistrationData](ConnectionValidator::Status status, const QStringList &errors) {
+                        if (OC_ENSURE(status == ConnectionValidator::Connected || status == ConnectionValidator::ServerVersionMismatch)) {
+                            // saving once after adding makes sure the account is stored in the config in a working state
+                            // this is needed to ensure a consistent state in the config file upon unexpected terminations of the client
+                            // (for instance, when running from a debugger and stopping the process from there)
+                            AccountManager::instance()->save(true);
+
+                            // only now, we can store the dynamic registration data in the keychain
+                            if (!dynamicRegistrationData.isEmpty()) {
+                                OAuth::saveDynamicRegistrationDataForAccount(accountStatePtr->account(), dynamicRegistrationData);
+                            }
+
+                            switch (syncMode) {
+                            case Wizard::SyncMode::SyncEverything:
+                            case Wizard::SyncMode::UseVfs: {
+                                bool useVfs = syncMode == Wizard::SyncMode::UseVfs;
+                                setUpInitialSyncFolder(accountStatePtr, useVfs);
+
+                                break;
+                            }
+                            case Wizard::SyncMode::ConfigureUsingFolderWizard: {
+                                Q_ASSERT(!accountStatePtr->account()->hasDefaultSyncRoot());
+
+                                auto *folderWizard = new FolderWizard(accountStatePtr, ocApp()->gui()->settingsDialog());
+                                folderWizard->resize(ocApp()->gui()->settingsDialog()->sizeHintForChild());
+                                folderWizard->setAttribute(Qt::WA_DeleteOnClose);
+
+                                // TODO: duplication of AccountSettings
+                                // adapted from AccountSettings::slotFolderWizardAccepted()
+                                connect(folderWizard, &QDialog::accepted, [accountStatePtr, folderWizard]() {
+                                    FolderMan *folderMan = FolderMan::instance();
+
+                                    qCInfo(lcApplication) << "Folder wizard completed";
+                                    const auto config = folderWizard->result();
+
+                                    auto folder = folderMan->addFolderFromFolderWizardResult(accountStatePtr, config);
+
+                                    if (!config.selectiveSyncBlackList.isEmpty() && OC_ENSURE(folder && !config.useVirtualFiles)) {
+                                        folder->journalDb()->setSelectiveSyncList(SyncJournalDb::SelectiveSyncBlackList, config.selectiveSyncBlackList);
+
+                                        // The user already accepted the selective sync dialog. everything is in the white list
+                                        folder->journalDb()->setSelectiveSyncList(SyncJournalDb::SelectiveSyncWhiteList, { QLatin1String("/") });
+                                    }
+
+                                    folderMan->setSyncEnabled(true);
+                                    folderMan->scheduleAllFolders();
+                                });
+
+                                connect(folderWizard, &QDialog::rejected, []() {
+                                    qCInfo(lcApplication) << "Folder wizard cancelled";
+                                    FolderMan::instance()->setSyncEnabled(true);
+                                });
+
+                                folderWizard->open();
+                                ocApp()->gui()->raiseDialog(folderWizard);
+
+                                break;
+                            }
+                            default:
+                                Q_UNREACHABLE();
+                            }
+                        }
+                    });
+
+
+                    validator->checkServer();
+                } else {
+                    FolderMan::instance()->setSyncEnabled(true);
+                }
+
+                // make sure the wizard is cleaned up eventually
+                _wizardController->deleteLater();
+            });
+
+        // all we have to do is show the dialog...
+        _wizardController->window()->show();
+        // ... and bring it to the front
+        raiseDialog(_wizardController->window());
     }
-    raiseDialog(settingsDialog());
 }
 
 void ownCloudGui::setPauseOnAllFoldersHelper(bool pause)
 {
-    QList<AccountState *> accounts;
+    QList<AccountStatePtr> accounts;
     if (auto account = qvariant_cast<AccountStatePtr>(sender()->property(propertyAccountC))) {
-        accounts.append(account.data());
+        accounts.append(account);
     } else {
         for (const auto &a : AccountManager::instance()->accounts()) {
-            accounts.append(a.data());
+            accounts.append(a);
         }
     }
-    for (auto *f : FolderMan::instance()->map()) {
+    for (auto *f : FolderMan::instance()->folders()) {
         if (accounts.contains(f->accountState())) {
             f->setSyncPaused(pause);
             if (pause) {
@@ -1047,47 +1189,52 @@ void ownCloudGui::slotShowShareDialog(const QString &sharePath, const QString &l
         qCWarning(lcApplication) << "Could not open share dialog for" << localPath << "no responsible folder found";
         return;
     }
-
-    const auto accountState = folder->accountState();
-
-    SyncJournalFileRecord fileRecord;
-
-    bool resharingAllowed = true; // lets assume the good
-    if (folder->journalDb()->getFileRecord(file, &fileRecord) && fileRecord.isValid()) {
-        // check the permission: Is resharing allowed?
-        if (!fileRecord._remotePerm.isNull() && !fileRecord._remotePerm.hasPermission(RemotePermissions::CanReshare)) {
-            resharingAllowed = false;
-        }
-    }
-
-    // As a first approximation, set the set of permissions that can be granted
-    // either to everything (resharing allowed) or nothing (no resharing).
-    //
-    // The correct value will be found with a propfind from ShareDialog.
-    // (we want to show the dialog directly, not wait for the propfind first)
-    SharePermissions maxSharingPermissions =
-        SharePermissionRead
-        | SharePermissionUpdate | SharePermissionCreate | SharePermissionDelete
-        | SharePermissionShare;
-    if (!resharingAllowed) {
-        maxSharingPermissions = SharePermission(0);
-    }
-
-
-    ShareDialog *w = nullptr;
-    if (_shareDialogs.contains(localPath) && _shareDialogs[localPath]) {
-        qCInfo(lcApplication) << "Raising share dialog" << sharePath << localPath;
-        w = _shareDialogs[localPath];
+    if (folder->accountState()->account()->capabilities().filesSharing().sharing_roles) {
+        fetchPrivateLinkUrl(folder->accountState()->account(), folder->webDavUrl(), sharePath, this, [](const QUrl &url) {
+            Utility::openBrowser(url, nullptr);
+        });
     } else {
-        qCInfo(lcApplication) << "Opening share dialog" << sharePath << localPath << maxSharingPermissions;
-        w = new ShareDialog(accountState, sharePath, localPath, maxSharingPermissions, startPage, settingsDialog());
-        w->setAttribute(Qt::WA_DeleteOnClose, true);
+        const auto accountState = folder->accountState();
 
-        _shareDialogs[localPath] = w;
-        connect(w, &QObject::destroyed, this, &ownCloudGui::slotRemoveDestroyedShareDialogs);
+        SyncJournalFileRecord fileRecord;
+
+        bool resharingAllowed = true; // lets assume the good
+        if (folder->journalDb()->getFileRecord(file, &fileRecord) && fileRecord.isValid()) {
+            // check the permission: Is resharing allowed?
+            if (!fileRecord._remotePerm.isNull() && !fileRecord._remotePerm.hasPermission(RemotePermissions::CanReshare)) {
+                resharingAllowed = false;
+            }
+        }
+
+        // As a first approximation, set the set of permissions that can be granted
+        // either to everything (resharing allowed) or nothing (no resharing).
+        //
+        // The correct value will be found with a propfind from ShareDialog.
+        // (we want to show the dialog directly, not wait for the propfind first)
+        SharePermissions maxSharingPermissions =
+            SharePermissionRead
+            | SharePermissionUpdate | SharePermissionCreate | SharePermissionDelete
+            | SharePermissionShare;
+        if (!resharingAllowed) {
+            maxSharingPermissions = SharePermission(0);
+        }
+
+
+        ShareDialog *w = nullptr;
+        if (_shareDialogs.contains(localPath) && _shareDialogs[localPath]) {
+            qCInfo(lcApplication) << "Raising share dialog" << sharePath << localPath;
+            w = _shareDialogs[localPath];
+        } else {
+            qCInfo(lcApplication) << "Opening share dialog" << sharePath << localPath << maxSharingPermissions;
+            w = new ShareDialog(accountState, folder->webDavUrl(), sharePath, localPath, maxSharingPermissions, startPage, settingsDialog());
+            w->setAttribute(Qt::WA_DeleteOnClose, true);
+
+            _shareDialogs[localPath] = w;
+            connect(w, &QObject::destroyed, this, &ownCloudGui::slotRemoveDestroyedShareDialogs);
+        }
+        w->open();
+        raiseDialog(w);
     }
-    w->open();
-    raiseDialog(w);
 }
 
 void ownCloudGui::slotRemoveDestroyedShareDialogs()
