@@ -211,8 +211,7 @@ void ShareLinkWidget::getShares()
 
 void ShareLinkWidget::slotSharesFetched(const QList<QSharedPointer<Share>> &shares)
 {
-    const QString versionString = _account->serverVersion();
-    qCInfo(lcSharing) << versionString << "Fetched" << shares.count() << "shares";
+    qCInfo(lcSharing) << "Fetched" << shares.count() << "shares";
 
     // Select the share that was previously selected,
     // except if an explicit override was asked for
@@ -263,7 +262,7 @@ void ShareLinkWidget::slotSharesFetched(const QList<QSharedPointer<Share>> &shar
         table->setItem(row, 0, nameItem);
 
         auto dotdotdotButton = new QToolButton;
-        dotdotdotButton->setText("...");
+        dotdotdotButton->setText(QStringLiteral("..."));
         dotdotdotButton->setProperty(propertyShareC, QVariant::fromValue(linkShare));
         connect(dotdotdotButton, &QAbstractButton::clicked, this, &ShareLinkWidget::slotContextMenuButtonClicked);
         table->setCellWidget(row, 1, dotdotdotButton);
@@ -373,7 +372,7 @@ void ShareLinkWidget::slotShareSelectionChanged()
         // only if setting was successful, we want to show a placeholder afterwards to signalize success
         if (share && share->isPasswordSet()) {
             _ui->checkBox_password->setChecked(true);
-            _ui->lineEdit_password->setPlaceholderText("********");
+            _ui->lineEdit_password->setPlaceholderText(QStringLiteral("********"));
             _ui->lineEdit_password->setEnabled(true);
         }
         _ui->lineEdit_password->setText(QString());
@@ -487,7 +486,7 @@ void ShareLinkWidget::slotPasswordSet()
     _ui->checkBox_password->setEnabled(true);
     _ui->lineEdit_password->setText(QString());
     if (share->isPasswordSet()) {
-        _ui->lineEdit_password->setPlaceholderText("********");
+        _ui->lineEdit_password->setPlaceholderText(QStringLiteral("********"));
         _ui->lineEdit_password->setEnabled(true);
     } else {
         _ui->lineEdit_password->setPlaceholderText(QString());
@@ -578,7 +577,7 @@ void ShareLinkWidget::slotCheckBoxExpireClicked()
 
 void ShareLinkWidget::emailShareLink(const QUrl &url)
 {
-    QString fileName = _sharePath.mid(_sharePath.lastIndexOf('/') + 1);
+    QString fileName = _sharePath.mid(_sharePath.lastIndexOf(QLatin1Char('/')) + 1);
     Utility::openEmailComposer(
         tr("I shared %1 with you").arg(fileName),
         url.toString(),
